﻿//
// RideRunner plug-in module for the HUD One
//

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.IO.Ports;
using System.Threading;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Timers;



/*
 * Remember every new plug in must have new GUIDS, there are 3:
 * The Interface class (IRRExtension.cs), the RRExtension class (RRExtsion.cs)
 * and in the properties for the ASSEMBLY (AssemblyInfo.cs)
*/

namespace HUDOneRR
{
    [Guid("EE9DD711-355A-4309-8162-A68B006903B0")]
	[ClassInterface(ClassInterfaceType.None)]

    class RRIndicator
    {
        bool state = false;      // TRUE IF THE INDICATOR IS LIT UP
        bool changed = false;    // TRUE IF THE STATE OF THIS INDICATOR 
                                 // HAS CHANGED SINCE LAST READ

        //================================================================
        public bool HasChanged()
        {

            return changed;
        }

        //================================================================
        public void Set (bool state)
        {
            lock (this)
            {
                this.state = state;
                this.changed = true;
            }
        }

        //================================================================
        public bool Get ()
        {
            bool state;

            lock (this)
            {
                state = this.state;
                this.changed = false;
            }
            return state;
        }

        //================================================================
        public override string ToString()
        {
            return state.ToString();
        }
    }

    class RRVar
    {
        int state = 0;           // CURRENT STATE OF THE VARIABLE
        bool changed = false;    // TRUE IF THE STATE OF THIS INDICATOR 
                                 // HAS CHANGED SINCE LAST READ

        //================================================================
        public bool HasChanged()
        {

            return changed;
        }

        //================================================================
        public void Set (int state)
        {
            lock (this)
            {
                this.state = state;
                this.changed = true;
            }
        }

        //================================================================
        public int Get ()
        {
            int state;

            lock (this)
            {
                state = this.state;
                this.changed = false;
            }
            return state;
        }

        //================================================================
        public override string ToString()
        {
            return state.ToString();
        }
    }

    class RRLabel
    {
        string contents = "";
        bool changed = false;

        //================================================================
        public bool HasChanged ()
        {

            return changed;
        }

        //================================================================
        public void Set (string str)
        {
            lock (this)
            {
                this.contents = str;
                this.changed = true;
            }
        }

        //================================================================
        public void Get (ref string str)
        {

            lock (this)
            {
                str = this.contents;
                this.changed = false;
            }
        }

        //================================================================
        public override string ToString()
        {
            return contents;
        }
    }

    public delegate void ChangedEventHandler(object sender, EventArgs e);

    class HUDOne
    {
        const int STAGEBUFSIZE = 512;
        const int HDRSIZE = 4;
        const byte MSG_SENTINEL = 0xff;
        byte[] m_stagebuf;
        byte[] m_sendbuf;
        int m_buflen = 0;
        Thread m_reader;
        SerialPort m_serial;
        bool m_shutdown = false;
        ulong m_flags = 0;                  // ACCUMULATED CHANGE FLAGS

        public V1STATUS v1;
        public CISTATUS ci;

        // MESSAGE IDS
        const byte MSG_PROTOCOL_VERSION = 1;
        const byte MSG_CLASS2 = 2;
        const byte MSG_V1_STATUS = 3;
        const byte MSG_CI_STATUS = 4;
        const byte MSG_V1_MUTE = 5;
        const byte MSG_CI_MUTE = 6;

        const byte MSG_REPLY_FLAG = 0x80;

        // V1 EVENT TAGS
        const byte VT_ALERT = 0;
        const byte VT_MODE = 1;
        const byte VT_ARROWS = 2;
        const byte VT_PRIORITY_ARROWS = 3;
        const byte VT_BANDS = 4;
        const byte VT_PRIORITY_BANDS = 5;
        const byte VT_BOGEY_COUNT = 6;
        const byte VT_RSSI = 7;
        const byte VT_MUTE = 8;
        const byte VT_JUNK = 9;
        const byte VT_ACTIVE = 10;
        const byte VT_NUM_TAGS = 11;

        // 9500CI EVENT TAGS
        const byte CT_ALERT = 0;
        const byte CT_ACTIVE = 1;
        const byte CT_MODE = 2;
        const byte CT_SENSITIVITY = 3;
        const byte CT_STD_BAND = 4;
        const byte CT_STD_RSSI = 5;
        const byte CT_EXP_ALERT = 6;
        const byte CT_SPEC_FREQ = 7;
        const byte CT_MARKER_BRG = 8;
        const byte CT_MARKER_DIST = 9;
        const byte CT_SPEED = 10;
        const byte CT_VOLTAGE = 11;
        const byte CT_VOLUME = 12;
        const byte CT_BRIGHTNESS = 13;
        const byte CT_STRING_TYPE = 14;
        const byte CT_STRING = 15;
        const byte CT_GPS_STAT = 16;
        const byte CT_NUM_TAGS = 17;

        // V1 CHANGE FLAGS
        public const ulong CF_V1_ALERT = 0x00000001;
        public const ulong CF_V1_ACTIVE = 0x00000002;
        public const ulong CF_V1_MODE = 0x00000004;
        public const ulong CF_V1_ARROWS = 0x00000008;
        public const ulong CF_V1_PRIORITY_ARROWS = 0x00000010;
        public const ulong CF_V1_BANDS = 0x00000020;
        public const ulong CF_V1_PRIORITY_BANDS = 0x00000040;
        public const ulong CF_V1_BOGEY_COUNT = 0x00000080;
        public const ulong CF_V1_RSSI = 0x00000100;
        public const ulong CF_V1_MUTE = 0x00000200;
        public const ulong CF_V1_JUNK = 0x00000400;

        // 9500CI CHANGE FLAGS
        public const ulong CF_CI_ALERT = 0x00000800;
        public const ulong CF_CI_ACTIVE = 0x00001000;
        public const ulong CF_CI_MODE = 0x00002000;
        public const ulong CF_CI_SENSITIVITY = 0x00004000;
        public const ulong CF_CI_STD_BAND = 0x00008000;
        public const ulong CF_CI_STD_RSSI = 0x00010000;
        public const ulong CF_CI_EXP_ALERT = 0x00020000;
        public const ulong CF_CI_SPEC_FREQ = 0x00040000;
        public const ulong CF_CI_MARKER_BRG = 0x00080000;
        public const ulong CF_CI_MARKER_DIST = 0x00100000;
        public const ulong CF_CI_SPEED = 0x00200000;
        public const ulong CF_CI_VOLTAGE = 0x00400000;
        public const ulong CF_CI_VOLUME = 0x00800000;
        public const ulong CF_CI_BRIGHTNESS = 0x01000000;
        public const ulong CF_CI_STRING_TYPE = 0x02000000;
        public const ulong CF_CI_STRING = 0x04000000;
        public const ulong CF_CI_GPS_STAT = 0x08000000;

        public event ChangedEventHandler Changed;

        //================================================================
        public HUDOne (string port)
        {

            m_sendbuf = new byte[STAGEBUFSIZE];
            m_stagebuf = new byte[STAGEBUFSIZE];
            m_reader = new Thread(Reader);
            m_reader.Name = "HUD One Reader Thread";

            v1 = new V1STATUS();
            ci = new CISTATUS();

            m_serial = new SerialPort(port, 115200, Parity.None, 8, StopBits.One);
            m_serial.Handshake = Handshake.None;
            m_serial.ReadTimeout = 2000;
        }

        //================================================================
        protected virtual void OnChanged (EventArgs e)
        {

            if (Changed != null)
            {
                Changed(this, e);
            }
        }

        //================================================================
        public ulong GetChangedFlags ()        
        {
            ulong flags;

            lock (this)
            {
                flags = m_flags;
                m_flags = 0U;
            }
            return flags;
        }

        //================================================================
        void AccumChangedFlags (ulong flags)
        {

            lock (this)
            {
                m_flags |= flags;
            }
        }

        //================================================================
        public bool Startup()
        {
            try
            {
                m_serial.Open();
            }
            catch (InvalidOperationException)
            {
                return false;
            }

            m_shutdown = false;
            m_reader.Start();

            // REQUEST THE CURRENT OPERATING MODES AND FLAGS
            byte[] b = new byte[3];
            b[0] = VT_MODE;
            b[1] = VT_ACTIVE;
            b[2] = VT_ALERT;
            SendMsg(MSG_V1_STATUS, 1, b, 3);

            b[0] = CT_SENSITIVITY;
            b[1] = CT_ACTIVE;
            b[2] = CT_ALERT;
            SendMsg(MSG_CI_STATUS, 2, b, 3);

            return true;
        }

        //================================================================
        public void Shutdown()
        {
            m_shutdown = true;
            m_reader.Join();
            m_serial.Close();
        }

        //================================================================
        void Dispatch (byte msgid, byte reqid, byte[] msgbuf, int offset, int len)
        {
            int curr = offset;
            int last = offset + len;

            switch (msgid & ~MSG_REPLY_FLAG)
            {
                case MSG_V1_STATUS:
                    while (curr < last)
                    {
                        switch (msgbuf[curr++])
                        {
                            case VT_ALERT:
                                v1.alert = (msgbuf[curr++] != 0);
                                AccumChangedFlags(CF_V1_ALERT);
                            break;

                            case VT_ACTIVE:
                                v1.active = (msgbuf[curr++] != 0);
                                AccumChangedFlags(CF_V1_ACTIVE);
                            break;

                            case VT_MUTE:
                                v1.mute = (msgbuf[curr++] != 0);
                                AccumChangedFlags(CF_V1_MUTE);
                            break;

                            case VT_JUNK:
                                v1.junk = (msgbuf[curr++] != 0);
                                AccumChangedFlags(CF_V1_JUNK);
                            break;

                            case VT_MODE:
                                v1.mode = msgbuf[curr++];
                                AccumChangedFlags(CF_V1_MODE);
                            break;

                            case VT_ARROWS:
                                v1.arrows = msgbuf[curr++];
                                AccumChangedFlags(CF_V1_ARROWS);
                            break;

                            case VT_PRIORITY_ARROWS:
                                v1.priority_arrows = msgbuf[curr++];
                                AccumChangedFlags(CF_V1_PRIORITY_ARROWS);
                            break;

                            case VT_BANDS:
                                v1.bands = msgbuf[curr++];
                                AccumChangedFlags(CF_V1_BANDS);
                            break;

                            case VT_PRIORITY_BANDS:
                                v1.priority_bands = msgbuf[curr++];
                                AccumChangedFlags(CF_V1_PRIORITY_BANDS);
                            break;

                            case VT_BOGEY_COUNT:
                                v1.bogey_count = msgbuf[curr++];
                                AccumChangedFlags(CF_V1_BOGEY_COUNT);
                            break;

                            case VT_RSSI:
                                v1.rssi = msgbuf[curr++];
                                AccumChangedFlags(CF_V1_RSSI);
                            break;

                            default:
                                curr = last;      // BAIL
                            break;
                        }
                    }
                    OnChanged(EventArgs.Empty);
                break;

                case MSG_CI_STATUS:
                    while (curr < last)
                    {
                        switch (msgbuf[curr++])
                        {
                            case CT_ALERT:
                                ci.alert = (msgbuf[curr++] != 0);
                                AccumChangedFlags(CF_CI_ALERT);
                            break;

                            case CT_ACTIVE:
                                ci.active = (msgbuf[curr++] != 0);
                                AccumChangedFlags(CF_CI_ACTIVE);
                            break;

                            case CT_MODE:
                                ci.pilot_mode = (CISTATUS.CI_PILOT_MODE)msgbuf[curr++];
                                AccumChangedFlags(CF_CI_MODE);
                            break;

                            case CT_SENSITIVITY:
                                ci.sensitivity = (CISTATUS.CI_SENSITIVITY)msgbuf[curr++];
                                AccumChangedFlags(CF_CI_SENSITIVITY);
                            break;

                            case CT_STD_BAND:
                                ci.band = (CISTATUS.CI_STD_BAND)msgbuf[curr++];
                                AccumChangedFlags(CF_CI_STD_BAND);
                            break;

                            case CT_STD_RSSI:
                                ci.rssi = msgbuf[curr++];
                                AccumChangedFlags(CF_CI_STD_RSSI);
                            break;

                            case CT_EXP_ALERT:
                                ci.exp = (ulong)msgbuf[curr++] << 24;
                                ci.exp |= (ulong)msgbuf[curr++] << 16;
                                ci.exp |= (ulong)msgbuf[curr++] << 8;
                                ci.exp |= (ulong)msgbuf[curr++];
                                AccumChangedFlags(CF_CI_EXP_ALERT);
                            break;

                            case CT_SPEC_FREQ:
                                ci.freq = (ulong)msgbuf[curr++] << 8;
                                ci.freq |= (ulong)msgbuf[curr++];
                                AccumChangedFlags(CF_CI_SPEC_FREQ);
                            break;

                            case CT_MARKER_BRG:
                                ci.brg = msgbuf[curr++];
                                AccumChangedFlags(CF_CI_MARKER_BRG);
                            break;

                            case CT_MARKER_DIST:
                                ci.dist = (ulong)msgbuf[curr++] << 8;
                                ci.dist |= (ulong)msgbuf[curr++];
                                AccumChangedFlags(CF_CI_MARKER_DIST);
                            break;

                            case CT_SPEED:
                                ci.speed = msgbuf[curr++];
                                AccumChangedFlags(CF_CI_SPEED);
                            break;

                            case CT_VOLTAGE:
                                ci.voltage = msgbuf[curr++];
                                AccumChangedFlags(CF_CI_VOLTAGE);
                            break;

                            case CT_VOLUME:
                                ci.volume = msgbuf[curr++];
                                AccumChangedFlags(CF_CI_VOLUME);
                            break;

                            case CT_BRIGHTNESS:
                                ci.brightness = msgbuf[curr++];
                                AccumChangedFlags(CF_CI_BRIGHTNESS);
                            break;

                            case CT_STRING_TYPE:
                                ci.strtype = (CISTATUS.CI_STRING_TYPE)msgbuf[curr++];
                                AccumChangedFlags(CF_CI_STRING_TYPE);
                            break;

                            case CT_STRING:
                                ci.str = System.Text.Encoding.ASCII.GetString(msgbuf, curr, 8);
                                curr += 8;
                                AccumChangedFlags(CF_CI_STRING);
                            break;

                            case CT_GPS_STAT:
                                ci.gps_stat = (CISTATUS.CI_GPS_STATUS)msgbuf[curr++];
                                AccumChangedFlags(CF_CI_GPS_STAT);
                            break;

                            default:
                                curr = last;      // BAIL
                            break;
                        }
                    }
                    OnChanged(EventArgs.Empty);
                break;

                default:
                break;
            }
        }

        //================================================================
        bool SendMsg (byte msgid, byte reqid, byte[] buf, byte len)
        {
            bool result = true;
            int offset = 0;

            lock(this)
            {
                // PREPARE THE HEADER
                m_sendbuf[offset++] = MSG_SENTINEL;
                m_sendbuf[offset++] = reqid;
                m_sendbuf[offset++] = msgid;
                m_sendbuf[offset++] = len;

                // ADD THE PAYLOAD
                Buffer.BlockCopy(buf, 0, m_sendbuf, offset, len);
                offset += len;

                // TACK ON THE CRC
                ushort crc = CRC.Calculate(m_sendbuf, 0, offset);
                m_sendbuf[offset++] = (byte)(crc & 0xff);
                m_sendbuf[offset++] = (byte)(crc >> 8);

                try 
                {
                    m_serial.Write(m_sendbuf, 0, offset);
                }
                catch (IOException)
                {
                    result = false;
                }
            }

            return result;
        }

        //================================================================
        bool ProcessNextMessage ()
        {
            int msglen = 0;
            int sentinel = 0;
            byte reqid = 0;
            byte msgid = 0;
            byte length = 0;

            m_buflen = 0;

            try
            {
                // ALIGN TO THE START OF A MESSAGE FRAME
                do
                {
                    sentinel = m_serial.ReadByte();
                } while (sentinel != MSG_SENTINEL);

                m_stagebuf[m_buflen++] = (byte)sentinel;

                // GET THE REST OF THE HEADER
                while (m_buflen < HDRSIZE)
                {
                    m_buflen += m_serial.Read(m_stagebuf, m_buflen, HDRSIZE - m_buflen);
                }
          
                reqid = m_stagebuf[1];
                msgid = m_stagebuf[2];
                length = m_stagebuf[3];
                msglen = HDRSIZE + length + sizeof(ushort);

                // GET THE REST OF THE MESSAGE
                while (m_buflen < msglen)
                {
                    m_buflen += m_serial.Read(m_stagebuf, m_buflen, msglen - m_buflen);
                }
            }
            catch (IOException)
            {
                return false;
            }
            catch (TimeoutException)
            {
                return false;
            }

            if (m_buflen >= msglen)
            {
                // DO WE HAVE A VALID MESSAGE?
                ushort crc = CRC.Calculate(m_stagebuf, 0, HDRSIZE + length);
                ushort crc_lo = m_stagebuf[msglen - 2];
                ushort crc_hi = m_stagebuf[msglen - 1];
                ushort crc_expected = (ushort)(crc_hi << 8);
                crc_expected |= crc_lo;
                if (crc == crc_expected)
                {
                    Dispatch(msgid, reqid, m_stagebuf, HDRSIZE, length);
                }
                return true;
            }
            return false;
        }

        //================================================================
        void Reader ()
        {
            while (!m_shutdown)
            {
                ProcessNextMessage();
            }
            SDK.Execute("EXIT");
        }
    }

    public class RRExtension : HUDOneRR.IRRExtension
	{
		private enum ProcessCommandResults
		{
			CommandNotProcessed = 0,
			CommandProcessedReturnPrevScreen = 1,
			CommandProcessedStayOnScreen = 2,
			CommandModified = 3
		}

		#region Private Fields

        bool m_pinned;                                  // TRUE IF OUR SCREEN IS (ALWAYS VISIBLE)
        bool m_visible;                                 // TRUE IF OUR SCREEN IS CURRENTLY VISIBLE
        bool m_active;                                  // TRUE IF ANY DISPLAY IS CURRENTLY ACTIVE
        Dictionary<string, RRIndicator> indmap;         // CURRENT STATUS OF INDICATORS
        Dictionary<string, RRLabel> labmap;             // CURRENT STATUS OF LABELS
        Dictionary<string, RRVar> varmap;               // CURRENT STATES OF VARIABLES
        HUDOne m_h;                                     // THE HUD ONE OBJECT
        System.Timers.Timer m_timer;                    // SCREEN TIMER

		#endregion

		#region Constructor

		public RRExtension ()
		{
            m_pinned = false;
            m_visible = false;
            m_active = false;

            m_timer = new System.Timers.Timer();
            m_timer.AutoReset = false;
            m_timer.Interval = 2500;                    // LINGER FOR 2.5 SECONDS
            m_timer.Elapsed += new ElapsedEventHandler(OnTimeout);

            indmap = new Dictionary<string, RRIndicator>();
            labmap = new Dictionary<string, RRLabel>();
            varmap = new Dictionary<string, RRVar>();

            // ## PULL THIS INFO FROM THE REGISTRY
            m_h = new HUDOne("COM15");
            m_h.Changed += new ChangedEventHandler(OnChanged);

            // POPULATE THE INDICATOR MAP
            indmap["!v1_front"] = new RRIndicator();
            indmap["!v1_sides"] = new RRIndicator();
            indmap["!v1_rear"] = new RRIndicator();
            indmap["!v1_l"] = new RRIndicator();
            indmap["!v1_ka"] = new RRIndicator();
            indmap["!v1_k"] = new RRIndicator();
            indmap["!v1_x"] = new RRIndicator();
            indmap["!v1_s1"] = new RRIndicator();
            indmap["!v1_s2"] = new RRIndicator();
            indmap["!v1_s3"] = new RRIndicator();
            indmap["!v1_s4"] = new RRIndicator();
            indmap["!v1_s5"] = new RRIndicator();
            indmap["!v1_s6"] = new RRIndicator();
            indmap["!v1_s7"] = new RRIndicator();
            indmap["!v1_s8"] = new RRIndicator();
            indmap["!ci_brg_0"] = new RRIndicator();
            indmap["!ci_brg_45"] = new RRIndicator();
            indmap["!ci_brg_90"] = new RRIndicator();
            indmap["!ci_brg_135"] = new RRIndicator();
            indmap["!ci_brg_180"] = new RRIndicator();
            indmap["!ci_brg_225"] = new RRIndicator();
            indmap["!ci_brg_270"] = new RRIndicator();
            indmap["!ci_brg_315"] = new RRIndicator();

            // POPULATE THE VARIABLE MAP
            varmap["ci_gps"] = new RRVar();

            // POPULATE THE LABEL MAP
            labmap["!v1_mode"] = new RRLabel();
            labmap["!v1_bogey_cnt"] = new RRLabel();
            labmap["!ci_disp"] = new RRLabel();
            labmap["!ci_mark"] = new RRLabel();

            if (!SDK.Created)
            {
                SDK.SetSDK();
            }
		}

		#endregion

		#region Public RRExtension API Methods (new ones)

		// This is called when the pluginmgr;about command is used
		//
		public void About(object frm)
		{
		}

		// This is called when the pluginmgr;settings command is used
		//
		public void Settings(object frm)
		{
		}

		#endregion

		#region Public RRExtension API Methods

		/// <summary>
		/// This is called ONLY once, right after plug-in is loaded
		/// use this to set up anything you need, including data storage
		/// </summary>
		/// <param name="pluginDataPath"></param>
		public void Initialize(string pluginDataPath)
		{
			/*
				pluginDataPath will contain a USER Profile (my documents) folder path
				suitable for storing WRITEABLE settings to
				this would make your plugin OS compliant (VISTA and onward)
				not to mention, its proper programming, user data should NOT be stored in "Program Files"
    		
			 * example (typical vista): "C:\Users\Username\Documents\RideRunner\Plugins\MyPlugin\"
    
				App.path will be the path of the ACTUALL LOADED .dll (not recomend for any writes)
			
				uncomment code below if u need the directory
			*/

			// use to create data storage folder
			//if (Directory.Exists(pluginDataPath) == false)
			//	Directory.CreateDirectory(pluginDataPath);
            m_h.Startup();
		}

		/// <summary>
		/// This is called when plugin manager is unloading plugin
		/// </summary>
		public void Terminate()
		{
			// clean up code goes here... 
			// close any file handles...free up any unmanaged resources
			// GC will take care of managed resources
            m_h.Shutdown();
		}

		/// <summary>
		/// called by plugin manager
		/// </summary>
		/// <param name="state"></param>
		public void Enabled(bool state)
		{
			SDK.Enabled = state;
		}

		/// <summary>
		/// called by plugin manager
		/// </summary>
		/// <param name="item"></param>
		/// <returns></returns>
		public string Properties(string item)
		{
			string properties = "";

			switch (item)
			{
				case "version":
					properties = Assembly.GetExecutingAssembly().GetName().Version.ToString();
				break;

				case "author":
					properties = "Swarga Research";
				break;

                case "category":
					properties = "TBD";
				break;

                case "description":
					properties = "HUD One interface";
				break;

                case "supporturl":
					properties = "http://www.swarga-research.com";
				break;
			}

			return properties;
		}

		//*****************************************************************
		//* This Function will be called with the current command string
		//* The return parameter of this function determines the handling
		//* To be taken upon returning to RR:
		//*
		//* 0 = Command not processed here
		//* 1 = Command completed + return to previous screen
		//* 2 = Command completed, stay on current screen
		//* 3 = Command has been changed/modified, execute returned one
		//*
		//* frm is the form object which generated the current command. Be
		//* VERY VERY careful when using it.
		//* must use InvokeMember InterOp on this object
		//*
		//* frm.tag contains the screen name for the screen the CMD was processed on.
		//*****************************************************************
		public int ProcessCommand(ref string cmd, object frm)
		{
			ProcessCommandResults result = ProcessCommandResults.CommandNotProcessed;

			// Example:
			//string frmTag = (string)frm.GetType().InvokeMember("Tag", BindingFlags.GetProperty, null, frm, null);

			switch (cmd.ToLower())
			{
				case "onsuspend":
                    m_h.Shutdown();
                    result = ProcessCommandResults.CommandProcessedStayOnScreen;
				break;

				case "onresume":
                    m_h.Startup();
                    result = ProcessCommandResults.CommandProcessedStayOnScreen;
				break;

                case "onscreenchange":
                    // UPDATE THE VISIBILITY FLAG
                    m_visible = (SDK.GetInfo("RRSCREEN") == "hudone.skin");
                    if (m_visible)
                    {
                        // FLUSH OUT ANYTHING THAT CHANGED WHILE WE WERE OFF
                        UpdateScreen();
                        if (!m_pinned && !m_active)
                        {
                            m_timer.Start();
                        }
                    }
                break;
			}

			return (int)result;
		}

		/// <summary>
		/// This Function will be called with a requested label code and
		/// format specified at the skin file. Simply return any text to
		/// be displayed for the specified format.
		/// </summary>
		/// <param name="lbl"></param>
		/// <param name="fmt"></param>
		/// <returns></returns>
		public string ReturnLabel (string lbl, string fmt)
		{
            string result = "";

            try
            {
                labmap[lbl].Get(ref result);
            }
            catch (KeyNotFoundException)
            {
                result = "";
            }
            return result;
		}

		// deprecated DONT USE
		public string ReturnIndicator (string Ind)
		{
            return "";
		}

		/// <summary>
		/// Called periodically with requested indicator code
		/// specified at the skin file. Simply return "True" or "False" to
		/// display the respective ON or OFF (EMPTY) layer of the skin images.
		/// alternatively you can specify a path to a file to be displayed
		/// as the indicator specified. Return "False" to erase the image.
		/// ONLY return something else IF AND ONLY IF you process that code
		/// </summary>
		/// <param name="Ind"></param>
		/// <returns>"" for not processed</returns>
		public string ReturnIndicatorEx (string ind)
		{
            string result;

            try
            {
                result = indmap[ind].ToString();
            }
            catch (KeyNotFoundException)
            {
                result = "";
            }

            return result;
		}

		/// <summary>
		/// This Sub will be called with an indicator code "CLICKED"
		/// specified at the skin file. This "event" so to speak can be used
		/// to toggle indicators or execute any code you desire when clicking
		/// on a specifig indicator in the skin. You can also modify IND and
		/// monitor the IND parameter as to detect/alter the behaviour of
		/// how RR will process the indicator code being clicked.
		/// </summary>
		/// <param name="IND"></param>
		public void IndicatorClick (ref string ind)
		{

			switch (ind.ToLower())
			{
				case "myind":
					SDK.ErrScrn("title", "myind", "i was clicked");
				break;
			}
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="SLD"></param>
		/// <returns>-1 for not processed</returns>
		public long ReturnSlider(string SLD)
		{
			return -1;
		}

		#endregion

		#region Logging

		[ComVisible(false)]
		static private void log(string message)
		{
			if (isAssemblyDebugBuild() == true)
			{
				SDK.DoLog(message);
			}
		}

		[ComVisible(false)]
		static private bool isAssemblyDebugBuild()
		{
			System.Reflection.Assembly assemb = System.Reflection.Assembly.GetExecutingAssembly();

			foreach (object att in assemb.GetCustomAttributes(false))
			{
				if (att.GetType() == System.Type.GetType("System.Diagnostics.DebuggableAttribute"))
				{
					return ((System.Diagnostics.DebuggableAttribute)att).IsJITTrackingEnabled;
				}
			}

			return false;
		}

		#endregion

		#region Private Methods

        //================================================================
        private void SetIndicator (string ind, bool state)
        {

            try
            {
                indmap[ind].Set(state);
            }
            catch (KeyNotFoundException)
            {

            }
        }

        ///================================================================
        ///<summary>
        ///Flushes all changed display elements out to the screen.
        ///</summary>
        ///================================================================
        private void UpdateScreen ()
        {

            if (m_visible)
            {
                string cmdstr;

                // UPDATE THE SCREEN TO REFLECT THE NEW INDICATOR STATES
                foreach (KeyValuePair<string,RRIndicator> ind in indmap)
                {
                    // IF THE DIRTY BIT HAS BEEN SET...
                    if (ind.Value.HasChanged())
                    {
                        // CONSTRUCT THE COMMAND STRING
                        cmdstr = "SETIND;" + ind.Key + ";" + (ind.Value.Get() ? "ON" : "OFF");

                        // AND EXECUTE THE COMMAND
                        SDK.Execute(cmdstr);
                    }
                }

                // UPDATE THE VARS
                foreach (KeyValuePair<string,RRVar> var in varmap)
                {
                    // IF THE DIRTY BIT HAS BEEN SET...
                    if (var.Value.HasChanged())
                    {
                        // CONSTRUCT THE COMMAND STRING
                        cmdstr = "SETVAR;" + var.Key + ";" + var.Value.ToString();

                        // AND EXECUTE THE COMMAND
                        SDK.Execute(cmdstr);
                    }
                }

                // UPDATE THE LABELS
                foreach (KeyValuePair<string,RRLabel> lab in labmap)
                {
                    // IF THE DIRTY BIT HAS BEEN SET...
                    if (lab.Value.HasChanged())
                    {
                        string contents = "";

                        // CONSTRUCT THE COMMAND STRING
                        lab.Value.Get(ref contents);
                        cmdstr = "SETLABEL;" + lab.Key + ";" + contents;

                        // AND EXECUTE THE COMMAND
                        SDK.Execute(cmdstr);
                    }
                }
            }
        }

        //================================================================
        private void OnTimeout (object source, ElapsedEventArgs e)
        {

            if (m_visible && !m_pinned)
            {
                SDK.Execute("EXIT");        // RETURN TO PREVIOUS SCREEN
            }
        }

        ///================================================================
        ///<summary>
        ///Updates screen elements to reflect changes in the current state(s)
        ///of the RADAR detector(s).
        ///</summary>
        ///================================================================
        private void OnChanged (object source, EventArgs e)
        {
            ulong flags = m_h.GetChangedFlags();

            // DETERMINE IF THE SCREEN NEEDS TO BE BROUGHT UP
            m_active = (m_h.v1.alert || m_h.v1.active || m_h.ci.alert || m_h.ci.active);

            if (m_active)
            {
                if (!m_visible)
                {
                    // BRING UP THE SCREEN
                    SDK.Execute("LOAD;hudone.skin");
                }
                m_timer.Stop();
            }

            // UPDATE MODE
            if ((flags & HUDOne.CF_V1_MODE) != 0)
            {
                string modestr = "";

                switch (m_h.v1.mode)
                {
                    case V1STATUS.MODE_ALL_BOGEYS:
                        modestr = "All Bogeys";
                    break;

                    case V1STATUS.MODE_LOGIC:
                        modestr = "Logic";
                    break;

                    case V1STATUS.MODE_ADVANCED_LOGIC:
                        modestr = "Advanced Logic";
                    break;
                }                
                labmap["!v1_mode"].Set(modestr);
            }

            // UPDATE ARROW INDICATORS
            if ((flags & HUDOne.CF_V1_ARROWS) != 0)
            {
                SetIndicator("!v1_rear", ((m_h.v1.arrows & V1STATUS.AF_REAR) == V1STATUS.AF_REAR));
                SetIndicator("!v1_sides", ((m_h.v1.arrows & V1STATUS.AF_SIDES) == V1STATUS.AF_SIDES));
                SetIndicator("!v1_front", ((m_h.v1.arrows & V1STATUS.AF_FRONT) == V1STATUS.AF_FRONT));
            }

            // UPDATE BAND INDICATORS
            if ((flags & HUDOne.CF_V1_BANDS) != 0)
            {
                SetIndicator("!v1_l", ((m_h.v1.bands & V1STATUS.BF_LASER) == V1STATUS.BF_LASER));
                SetIndicator("!v1_ka", ((m_h.v1.bands & V1STATUS.BF_KABAND) == V1STATUS.BF_KABAND));
                SetIndicator("!v1_k", ((m_h.v1.bands & V1STATUS.BF_KBAND) == V1STATUS.BF_KBAND));
                SetIndicator("!v1_x", ((m_h.v1.bands & V1STATUS.BF_XBAND) == V1STATUS.BF_XBAND));
            }

            // UPDATE RECEIVED SIGNAL STRENGTH INDICATORS
            if ((flags & HUDOne.CF_V1_BANDS) != 0)
            {
                SetIndicator("!v1_s1", (m_h.v1.rssi >= 1));
                SetIndicator("!v1_s2", (m_h.v1.rssi >= 2));
                SetIndicator("!v1_s3", (m_h.v1.rssi >= 3));
                SetIndicator("!v1_s4", (m_h.v1.rssi >= 4));
                SetIndicator("!v1_s5", (m_h.v1.rssi >= 5));
                SetIndicator("!v1_s6", (m_h.v1.rssi >= 6));
                SetIndicator("!v1_s7", (m_h.v1.rssi >= 7));
                SetIndicator("!v1_s8", (m_h.v1.rssi >= 8));
            }

            // UPDATE BOGEY COUNT
            if ((flags & HUDOne.CF_V1_BOGEY_COUNT) != 0)
            {
                string count = "";

                if (m_h.v1.bogey_count > 0)
                {
                    count = m_h.v1.bogey_count.ToString();
                }
                labmap["!v1_bogey_cnt"].Set(count);
            }

            // UPDATE 9500CI DISPLAY
            if ((flags & HUDOne.CF_CI_GPS_STAT) != 0)
            {
                varmap["ci_gps"].Set((int)m_h.ci.gps_stat);
            }

            if ((flags & HUDOne.CF_CI_STRING) != 0)
            {
                if (m_h.ci.strtype == CISTATUS.CI_STRING_TYPE.ST_INVISIBLE)
                {
                    labmap["!ci_disp"].Set("");
                }
                else
                {
                    labmap["!ci_disp"].Set(m_h.ci.str.Trim());
                }
            }

            if ((flags & HUDOne.CF_CI_SPEC_FREQ) != 0)
            {
                double ghz = (double)m_h.ci.freq / 1000.0;
                string alert = "";

                // THE BAND IS IMPLICIT IN THE FREQUENCY
                if (ghz > 30.0)
                {
                    // KA BAND
                    alert = "Ka ";
                }
                else if (ghz > 20.0)
                {
                    // K BAND
                    alert = "K ";
                }
                else
                {
                    // MUST BE X
                    alert = "X ";
                }
                alert += ghz.ToString("0.000");
                labmap["!ci_disp"].Set(alert);
            }

            // DISPLAY DISTANCE TO MARKED LOCATION
            if ((flags & HUDOne.CF_CI_MARKER_DIST) != 0)
            {
                string dist = "";

                dist = m_h.ci.dist.ToString();
                labmap["!ci_mark"].Set(dist);
            }

            // DISPLAY BEARING TO MARKED LOCATION
            if ((flags & HUDOne.CF_CI_MARKER_BRG) != 0)
            {
                ulong brg = (ulong)m_h.ci.brg >> 2;

                SetIndicator("!ci_brg_0", (brg == 0));
                SetIndicator("!ci_brg_45", (brg == 1));
                SetIndicator("!ci_brg_90", (brg == 2));
                SetIndicator("!ci_brg_135", (brg == 3));
                SetIndicator("!ci_brg_180", (brg == 4));
                SetIndicator("!ci_beg_225", (brg == 5));
                SetIndicator("!ci_brg_270", (brg == 6));
                SetIndicator("!ci_brg_315", (brg == 7));
            }

            if ((flags & (HUDOne.CF_CI_ALERT | HUDOne.CF_CI_ACTIVE)) != 0)
            {
                if (!m_h.ci.alert && !m_h.ci.active)
                {
                    string modestr = "";

                    // EXITING - DISPLAY CURRENT SENSITIVITY SETTING
                    switch (m_h.ci.sensitivity)
                    {
                        case CISTATUS.CI_SENSITIVITY.S_HIGHWAY:
                            modestr = "Highway";
                        break;

                        case CISTATUS.CI_SENSITIVITY.S_AUTO:
                            modestr = "Auto";
                        break;

                        case CISTATUS.CI_SENSITIVITY.S_AUTO_NO_X:
                            modestr = "Auto No X";
                        break;
                    }
                    labmap["!ci_disp"].Set(modestr);

                    // CLEAR MARKED LOCATION DISPLAY
                    SetIndicator("!ci_brg_0", false);
                    SetIndicator("!ci_brg_45", false);
                    SetIndicator("!ci_brg_90", false);
                    SetIndicator("!ci_brg_135", false);
                    SetIndicator("!ci_brg_180", false);
                    SetIndicator("!ci_brg_225", false);
                    SetIndicator("!ci_brg_270", false);
                    SetIndicator("!ci_brg_315", false);
                    labmap["!ci_mark"].Set("");
                }
            }

            // FLUSH ALL LABELS AND INDICATORS TO THE SCREEN
            UpdateScreen();

            // DISMISS THE SCREEN IF ALL DISPLAYS ARE IDLE
            if (!m_pinned && m_visible && !m_active)
            {   
                m_timer.Start();
            }
        }
		#endregion
	}
}

